/*
    CSC714 Term Project

    Task Scheduling Simulator

    Group Info:

    sykang      Sangyeol    Kang
    kabhavsa    Kinjal      Bhavsar 
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include "sched.h"

//#define DEBUG_ENABLE
//#undef DEBUG_ENABLE

#ifdef DEBUG_ENABLE
#define DEBUG printf
#else
#define DEBUG 
#endif

// Array of induced tasks
tasks_t tasks[MAX_TASK];

// string to store messages
char str[32];

// flag to show missed deadline
int missDeadline = 0;

//declaring the current system state
state_t state = {0};

// systems parameters
unsigned int startTime = 0, stopTime = 0;

unsigned int startDisplay=0;
unsigned int endDisplay=0;

//task behavior history
int *HIST;

qnode qHead = NULL;

//graph color code in RGB format
const char GraphColor[MAX_TASK][8] = {
  "#FFFF00",  "#33FFFF",  "#99FF99",  "#6699FF",  "#00FFCC",
  "#CC00CC",  "#990099",  "#336699",  "#FFCCCC",  "#CCCCCC",
  "#009999",  "#99CCFF",  "#CCFF00",  "#660099",  "#FF00CC",
  "#993399",  "#CC9900",  "#FF9900",  "#006600",  "#666699",
  "#999966",  "#33CCFF",  "#FF9966",  "#CC66FF",  "#993300",
  "#660000",  "#FF6600",  "#CC6600",  "#996699",  "#FFCC00",
  "#CCCC00",  "#9900CC" };

int main(int argc, char *argv[])
{
    unsigned int i = 0;
    if (argc < 2)
    {
        printf("USAGE: .//sched <input file>");
        exit(-1);
    }

    if(init_simulator(argv[1])==FALSE)
	return FALSE;
    if(init_state()==FALSE) 
	return FALSE;

    //Simulation
    if(state.scheme==FIXED_PTS || state.scheme==RM_PTS || state.scheme==DM_PTS)
    {
      calcMPT();
      /*tasks[0].PT=1;
      tasks[1].PT=2;
      tasks[2].PT=3;*/

      init_state();
      schedule_taskset();
    }
    else schedule_taskset();

    //print out the given information
    printf("** Task Scheduling Simulator **\n");
    printf("- Simulated # of scheduler tics: %d\n", stopTime);
    printf("- Simulated scheduling scheme: %s\n", state.scheme_name);
    printf("+ Given task set: %s\n", argv[1]);
    printf("  TaskID    Phase   Period     WCET     BCET   Deadline   Task Name\n");
    for(i=0; i<state.num_tasks; i++)
    {
      printf("%8d %8d %8d %8d %8d   %8d   %s\n",
             tasks[i].ID, tasks[i].phase, tasks[i].period,
             tasks[i].wcet, tasks[i].bcet, tasks[i].deadline, tasks[i].name);
    }

    //print out the simulation result
    printf("+ Simulation statistics\n");
    printf("  TaskID    #jobs  #jobs done  #missed DL    Priority\n");
    for (i=0; i<state.num_tasks; i++)
    {
      printf("%8d %8d  %10d %11d    %8d   \n", tasks[i].ID, tasks[i].cInstance.num, tasks[i].cInstance.num - tasks[i].cInstance.missDeadline, 
        tasks[i].cInstance.missDeadline, tasks[i].priority);
    }

    if(state.scheme==FIXED_PTS || state.scheme==RM_PTS || state.scheme==DM_PTS)
    {
      printf("+ Simulated Preemption Threshold\n");
      for(i=0; i<state.num_tasks; i++) printf("  - Task %d => %d\n", tasks[i].ID, tasks[i].PT);      
    }

    printf("+ Simulated WCRT\n");
    for(i=0; i<state.num_tasks; i++) printf("  - Task %d - %d\n", tasks[i].ID, calcWCRT(i));
    
    //generate the final ouptuts    
    genOutput();
    
    free(HIST);
    
    return 0;
}

int init_simulator(char *fin)
{
  //initialize global variables
  state.hyperperiod = 0;
  state.sim_time = 0;
  state.scheme = FIXED;
  state.PTS = FALSE;
  state.preemptive = FALSE;

  memset (tasks, 0, sizeof(tasks_t)*MAX_TASK);

  if(parseInput(fin)==FALSE) 
      return FALSE;
      
  if(verifyInput()==FALSE) 
      return FALSE;

  calcPriority(state.scheme);

  stopTime = state.sim_time < state.hyperperiod ? state.sim_time : state.hyperperiod;
  
  //check upper limit of simulation time
  //if(stopTime>128) stopTime=128;
  
  //allocate array for stroing task behaviors
  HIST=(int*)malloc(sizeof(int)*state.num_tasks*(stopTime+1));
  memset(HIST, 0, sizeof(int)*state.num_tasks*(stopTime+1));    

  return TRUE;
}

int init_state(void)
{
    unsigned int i = 0;
    qnode tmp;
    memset(HIST, 0, sizeof(int)*state.num_tasks*(stopTime+1));    
    // Setting the task states
    for(i=0; i<state.num_tasks; i++)
    {
        tasks[i].cInstance.arrival = tasks[i].phase;
        tasks[i].cInstance.cDeadline = tasks[i].phase + tasks[i].deadline;
        tasks[i].exec = tasks[i].wcet;
    }
    // System priority is the lowest
    state.sys_prio = state.num_tasks + 1;
    while (qHead != NULL)
      {
        tmp = qHead;
        qHead = qHead->next;
        free(tmp);
      }
      qHead = NULL;

    // Setting start time
    state.sys_time = 0;    
    return TRUE;
}

void schedule_taskset()
{
    //scheduling
    DEBUG("Time Job Action\n");    
    do{
        //check_deadline();
        state_handler();
        scheduler();
        check_deadline();
        //state_handler();
        
        if (startTime <= state.sys_time)
        {
            DEBUG(" HIST: %d %d %d\n", HIST[state.sys_time*state.num_tasks], HIST[state.sys_time*state.num_tasks+1], HIST[state.sys_time*state.num_tasks+2]);
        }
        state.sys_time++;
    }while (stopTime >= state.sys_time); // Go upto endtime
}

int parseInput(char *fn)
{
  FILE *fptr;
  char temp[MAX_LEN_STR];
  int ret;
  int ind;
  char c;
  int i=0;

  fptr=fopen(fn, "r");
  if(fptr==NULL)
  {
    printf("- Error: Cannot open the given input file\n");
    return FALSE;
  }
  
  //initialize return value
  ret=TRUE;
  state.num_tasks=0;

  /*
  indicator notices the sequence of input stream
  ind==0; Nothing
  ind==1; Simulation Time
  ind==11; starting time for displaying window
  ind==12; end time for displaying window
  ind==2; Scheduling Scheme
  ind==22; PTS option
  ind==222; Preemptive option
  ind==3; Task information will be parsed
  ind==4; New task
  ind==5; Task ID done
  ind==6; Task Name done
  ind==7; Task phase done
  ind==8; Task period done
  ind==9; WCET
  ind==10; BCET
  ind==11; Task deadline done
  */

  ind=0;

  //scan stream
  while(1)
  {
    c=fscanf(fptr, "%s", temp);
    if(c==EOF && feof(fptr)) break;
    if(c==EOF && ferror(fptr))
    {
      printf("- Error occurs while reading input file\n");
      ret=0;
      break;
    }

    if(strcmp(temp, "Time=")==0) 
    {
      ind=1;
      DEBUG("**Time=** identified\n");
      continue;
    }

    if(strcmp(temp, "StartDisplay=")==0) 
    {
      ind=11;
      DEBUG("**StartDisplay=** identified\n");
      continue;
    }

    if(strcmp(temp, "EndDisplay=")==0) 
    {
      ind=111;
      DEBUG("**EndDisplay=** identified\n");
      continue;
    }

    if(strcmp(temp, "Scheme=")==0) 
    {
      ind=2;
      DEBUG("**Scheme=** identified\n");
      continue;
    }
    
    if(strcmp(temp, "PTS=")==0) 
    {
      ind=22;
      DEBUG("**PTS=** identified\n");
      continue;
    }

    if(strcmp(temp, "Preemptive=")==0) 
    {
      ind=222;
      DEBUG("**Preemptive=** identified\n");
      continue;
    }

    if(strcmp(temp, "Deadline")==0)
    {
      ind=3;
      DEBUG("**Deadline** identified\n");
    }

    switch(ind)
    {
      //Start
      case 1:
        state.sim_time=atoi(temp);
        ind=0;
        break;

      case 11:
        startDisplay=atoi(temp);
        ind=0;
        break;      

      case 111:
        endDisplay=atoi(temp);
        ind=0;
        break;      

      case 2:
        if(strcmp(temp, "RM")==0) state.scheme=RM;
        else if(strcmp(temp, "DM")==0) state.scheme=DM;
        else if(strcmp(temp, "EDF")==0) state.scheme=EDF;
        else if(strcmp(temp, "FIXED")==0) state.scheme=FIXED;
        else state.scheme=FIXED;
        ind=0;
        break;

      case 22:
        if(strcmp(temp, "Yes")==0) state.PTS=TRUE;
        else state.PTS=FALSE;
        ind=0;
        break;

      case 222:
        if(strcmp(temp, "Yes")==0) state.preemptive=TRUE;
        else state.preemptive=FALSE;
        ind=0;
        break;
        
      case 3:
        ind=4;
        break;

      case 4:
        tasks[state.num_tasks].ID=atoi(temp);
        ind=5;
        break;

      case 5:
        tasks[state.num_tasks].name=strdup(temp);
        ind=6;
        break;

      case 6:
        tasks[state.num_tasks].phase=atoi(temp);
        ind=7;
        break;

      case 7:
        tasks[state.num_tasks].period=atoi(temp);
        ind=8;
        break;

      case 8:
        tasks[state.num_tasks].wcet=atoi(temp);
        ind=9;
        break;

      case 9:
        tasks[state.num_tasks].bcet=atoi(temp);
        ind=10;
        break;

      case 10:
        tasks[state.num_tasks].deadline=atoi(temp);
        ind=4;
        state.num_tasks++;
        break;
      case 0:
      default:
        break;
    }
  }

  return ret;
}

int verifyInput(void)
{
  unsigned int i;
  unsigned int j;

  //check deadline is within period
  for(i=0; i<state.num_tasks; i++)
  {
    if(tasks[i].deadline > tasks[i].period) return FALSE;
  }  

  //determine final simulation state.scheme
  switch (state.scheme)
  {
    case EDF:
      state.PTS=FALSE;
      sprintf(state.scheme_name, "EDF");
      break;

    case RM:
      if(state.preemptive)
      {
        if(state.PTS)
        {
          state.scheme=RM_PTS;
          sprintf(state.scheme_name, "RM, PTS");
        }
        else
        {
          state.scheme=RM_P;
          sprintf(state.scheme_name, "RM, Preemptive");
        }
      }
      else
      {
        state.scheme=RM;
        sprintf(state.scheme_name, "RM, Nonpreemptive");
      }
      break;

    case DM:
      if(state.preemptive)
      {
        if(state.PTS)
        {
          state.scheme=DM_PTS;
          sprintf(state.scheme_name, "DM, PTS");
        }
        else
        {
          state.scheme=DM_P;
          sprintf(state.scheme_name, "DM, Preemptive");
        }
      }
      else
      {
        state.scheme=DM;
        sprintf(state.scheme_name, "DM, Nonpreemptive");
      }
      break;

    case FIXED:
    default:
      if(state.preemptive)
      {
        if(state.PTS)
        {
          state.scheme=FIXED_PTS;
          sprintf(state.scheme_name, "Fixed, PTS");
        }
        else
        {
          state.scheme=FIXED_P;
          sprintf(state.scheme_name, "Fixed, Preemptive");
        }
      }
      else
      {
        state.scheme=FIXED;
        sprintf(state.scheme_name, "Fixed, Nonpreemptive");
      }
      break;
  }
      
  //compute hyperperiod
  if(state.num_tasks==1) state.hyperperiod=tasks[0].period;
  else if(state.num_tasks==2) state.hyperperiod=LCM(tasks[0].period, tasks[1].period);
  else
  {
    j=LCM(tasks[0].period, tasks[1].period);
    for(i=2; i<state.num_tasks; i++)
    {
      j=LCM(j, tasks[i].period);
    }
    state.hyperperiod=j;
  }

  DEBUG("Hyper Period=%d\n", state.hyperperiod);

  return TRUE;
}

int GCD(int a, int b)
{
  int i;
  int j;
  int gcd;

  gcd=1;

  //loop until reach the larger value  
  if(a>=b) j=a;
  else j=b;
  
  for(i=2; i<j; i++)
  {
    //if we can divide two arguments commonly
    if(a%i==0 && b%i==0) gcd=i;
  }

  return gcd;
}


int LCM(int a, int b)
{
  int lcm;

  //by LCM=a*b / GCD(a,b)
  
  lcm=a*b / GCD(a, b);

  return lcm;
}
void addqueue(tasks_t cTask)
{
    qnode tmp, next;
    tmp = qHead;
    next = (qnode) malloc (sizeof (struct qnode_t));
    next->deadline = cTask.cInstance.cDeadline;
    next->priority = cTask.priority;
    next->pth = cTask.PT;
    next->left = cTask.exec;
    next->taskid = cTask.ID;
    next->state = READY;
    next->next = NULL;
    HIST[state.sys_time*state.num_tasks+next->taskid] |= READY;       // Storing the ready state in History buffer

//If readyq is empty, make it first node
    if (NULL == qHead)
    {
        qHead = next;
        return;
    }
    //if we are the highest priority node, insert at head
    if (next->priority < qHead->priority)
    {
        if (state.preemptive)
        {
            next->next = qHead;
            qHead = next;
            return;
        }
        else 
        {
            if (qHead->state == RUNNING)
            {
                next->next = qHead->next;
                qHead->next = next;
                return;
             }
            else
            {
                next->next = qHead;
                qHead = next;
                return;
            }
        }
    }
    tmp = qHead;
    //Finding the right place to insert according to 
    while (tmp->next != NULL)
    {
        if (next->priority < tmp->next->priority)
        {
            next->next = tmp->next;
            tmp->next = next;
            return;
        }
        tmp = tmp->next;
    }
    // The node does not fir anywhere, insert at end
    tmp->next = next;
    return;
}
//Function to check tasks and set them to ready on time
void state_handler(void)
{
    unsigned int i = 0, j = 0;
    tasks_t temp;

    if (state.scheme == EDF)
    {
        for (i=0; i<state.num_tasks; i++)
        {
            for (j=i+1; j<state.num_tasks; j++)
            {
                if (tasks[i].cInstance.cDeadline > tasks[i].cInstance.cDeadline)
                {
                    temp = tasks[i];
                    tasks[i] = tasks[j];
                    tasks[j] = temp;
                }
            }
        }
        for (i=0; i<state.num_tasks; i++)
        {
            tasks[i].priority = i+1;
        }
    }

    for (i=0; i<state.num_tasks; i++)
    {
        if (tasks[i].cInstance.arrival <= state.sys_time)   //if arrival time is same as system current time then task is ready
        {
            tasks[i].cInstance.state = READY;
            tasks[i].cInstance.num++;                  // increase the instance of the task
            tasks[i].cInstance.time_left = tasks[i].exec;   // putting the execution time left
            tasks[i].cInstance.cDeadline = state.sys_time + tasks[i].deadline;  // calculating the deadline for the current instance
            tasks[i].cInstance.arrival = state.sys_time + tasks[i].period;      // calculating the next time the task will become active
            if (tasks[i].cInstance.cDeadline < stopTime)
                HIST[tasks[i].cInstance.cDeadline*state.num_tasks+i] |= OVER;
            addqueue(tasks[i]);

        }
    }
}
qnode remqnode(qnode cnode)
{
    qnode tmp, prev;
    prev = NULL;
    tmp = qHead;
    
    while (tmp != NULL && tmp != cnode)
    {
        prev = tmp;
        tmp = tmp->next;
    }
    
    if (tmp != NULL)
    {
        //if tmp==head of the queue
        if (prev == NULL)
        {
            prev = cnode->next;
            qHead = prev;
        }
        else
        {
            prev->next = cnode->next;
        }

        //cnode->next=NULL;
        free(cnode);
        
        if (prev != NULL)
            return prev->next;
        else
            return NULL;
    }
    return NULL;
}

void check_deadline(void)
{
    unsigned i;
    qnode tmp, tmp2;
    
    tmp = qHead;
    
    while (tmp != NULL)
    {
        if (tmp->deadline <= state.sys_time)
        {
            DEBUG(" deadline for %d missed", tmp->taskid);
            //HIST[state.sys_time*state.num_tasks+qHead->taskid] |= OVER;
            tmp->state = OVER;
            tasks[tmp->taskid].cInstance.missDeadline++;
            tmp2 = tmp->next;
            remqnode(tmp);
            tmp = tmp2;
            continue;
        }
        tmp = tmp->next;
    }

#if 0
    for (i=0; i<state.num_tasks; i++)
    {
        if(tasks[i].cInstance.cDeadline == state.sys_time ) // Check for missed deadlines
        {
            HIST[state.sys_time*state.num_tasks+i]|=OVER;

            /*if(tasks[i].cInstance.time_left > 0 && tasks[i].cInstance.state != OVER)
            {
                //sprintf(str," deadline for %s%d missed", tasks[i].name, tasks[i].cInstance.num-1);
                DEBUG(" deadline for %s%d missed", tasks[i].name, tasks[i].cInstance.num-1);
                tasks[i].cInstance.state = OVER;
                tasks[i].cInstance.missDeadline++;
                //missDeadline = 1; //Set the flag to say deadline is over
            }*/
        }
    }
    #endif
}

void remqhead()
{
	qnode tmp;
	tmp = qHead;
	qHead = qHead->next;
	free(tmp);
}

void scheduler(void)
{
    qnode currjob = NULL;
    qnode tmp = qHead;
    while (tmp != NULL)
      {
        //printf("\nID Left %d %d\n", tmp->taskid, tmp->left);
        tmp = tmp->next;
      }
    if (qHead == NULL)
    {
        DEBUG("%4.2d",state.sys_time);
    }
    else
    {
        DEBUG("%4.2d %d ",state.sys_time, qHead->taskid);
        currjob = state.cjob;
        if (currjob != NULL )
        {
            if (currjob != qHead && currjob->left > 0)
            {   
                DEBUG(" preempting");
                currjob = qHead;
                state.cjob = qHead;
            }
        }
        if (currjob == NULL)
        {
            currjob = qHead;
            state.cjob = qHead;
        }
        qHead->state = RUNNING;
        qHead->priority = qHead->pth;

        HIST[state.sys_time*state.num_tasks+qHead->taskid] |= RUNNING;       // Storing the ready state in History buffer
        if (--qHead->left == 0 )
        {
            DEBUG(" DONE");
            remqhead();
            currjob = NULL;
            state.cjob = NULL;
        }
    }

}

int calcPriority(SCHEDULE_n type)
{
  unsigned int i, j;
  tasks_t tmp;
  tasks_t *dest;

  dest=(tasks_t*)malloc(sizeof(tasks_t)*state.num_tasks);
  for(i=0; i<state.num_tasks; i++)
  {
    dest[i]=tasks[i];
  }

  switch(type)
  {
    case RM:
    case RM_P:
    case RM_PTS:
        //initialize array with period
        for(i=0; i<state.num_tasks; i++)
        {
            for(j=i+1; j<state.num_tasks; j++)
            {
                if (dest[j].period < dest[i].period)
                {
                    tmp = dest[i];
                    dest[i] = dest[j];
                    dest[j] = tmp;
                }
            }
        }
        for(i=0; i<state.num_tasks; i++)
        {
            for(j=0; j<state.num_tasks; j++)
            {
              if(dest[j].ID==i)
              {              
              tasks[i].priority = j+1;
              tasks[i].PT = j+1;
              }
            }
        }

      DEBUG("> RM Prioritization\n");
      for(i=0; i<state.num_tasks; i++) 
        DEBUG("Pri %d - Task %s, Period %d\n", i, dest[i].name, dest[i].period);
      break;
      
    case DM:
    case DM_P:
    case DM_PTS:
        for(i=0; i<state.num_tasks; i++)
        {
            for(j=i+1; j<state.num_tasks; j++)
            {
                if (dest[i].deadline > dest[j].deadline)
                {
                    tmp = dest[j];
                    dest[j] = dest[i];
                    dest[i] = tmp;
                }
            }
        }
        for(i=0; i<state.num_tasks; i++)
        {
            for(j=0; j<state.num_tasks; j++)
            {
              if(dest[j].ID==i)
              {              
              tasks[i].priority = j+1;
              tasks[i].PT = j+1;
              }
            }
        }
      DEBUG("> DM Prioritization\n");
      for(i=0; i<state.num_tasks; i++) DEBUG("Pri %d - Task %s, Deadline %d\n", i, tasks[i].name, tasks[i].deadline);     
      break;

    case EDF:
          //there is no fixed priority
    case FIXED:
    case FIXED_P:
    case FIXED_PTS:
          //specified in priority order
    default:
      break;
  }

  free(dest);
  
  return TRUE;
}

int calcMPT(void)
{
  int i;
  int j;
  unsigned int e1;
  unsigned int e2;
  tasks_t *temp[state.num_tasks];

  //initialize priority with its original priority
  for(i=0; i<state.num_tasks; i++) tasks[i].PT=tasks[i].priority;

  for(i=0; i<state.num_tasks; i++)
    {
      for(j=0; j<state.num_tasks; j++)
      {
        if(tasks[j].priority-1==i)
        {
          temp[i]=&tasks[j];
          break;
        }
      }
    }
  
  //simulate default schedule
  schedule_taskset();
  e1=calcWCRT(1);
  
  for(i=1; i<state.num_tasks; i++)
  {
    j=i-1;
    while(e1<temp[i]->deadline && temp[i]->PT>=0 && j>=0)
    {
      //increase PT by 1 and examine
      temp[i]->PT--;
      //for(i=0; i<state.num_tasks; i++) printf ("temp[i]->PT, exec :%d %d",temp[i]->PT, temp[i]->exec);
      init_state();
      schedule_taskset();

      //check schedulability
      e2=calcWCRT(j);
      if(e2>temp[j]->deadline) temp[i]->PT++;

      e1=calcWCRT(i);
      j--;
    }
  }

  //free(temp);
  
  return TRUE;
}

int calcWCRT(int taskID)
{
  unsigned int i;
  int rel;
  int comp;
  int wcrt;
  int prev;
  int curr;

  //calculate the WCRT of the task having taskID within Hyperperiod  
  rel=0;
  comp=0;
  wcrt=0;
  prev=HIST[taskID];
  curr=prev;
  for(i=0; i<stopTime; i++)
  {
    curr=HIST[i*state.num_tasks+taskID];

    if(curr&READY)
    {
        if((comp-rel+1)>wcrt) wcrt=comp-rel+1;
        rel=i;
    }
        
    if(curr&RUNNING) comp=i;

    prev=curr;
  }

  return wcrt;
}

int genOutput(void)
{
    int i;
    int j;
    int end;
    char temp[80];
    
    FILE **fptrTask;
    FILE *fptr;
    
    fptrTask=(FILE**)malloc(sizeof(FILE*)*state.num_tasks);
    
    for(i=0; i<state.num_tasks; i++)
    {
        sprintf(temp, "TASK-%d.dat", tasks[i].ID);
        fptrTask[i]=fopen(temp, "wt");
    }

    if(endDisplay>stopTime) end=stopTime;
    else end=endDisplay;

    for(i=startDisplay; i<=end; i++)
    {
        for(j=0; j<state.num_tasks; j++)
        {
            fprintf(fptrTask[j], "%.1f\t", (i+0.5));
            if(HIST[i*state.num_tasks+j]&RUNNING) fprintf(fptrTask[j], "1\t");
            else fprintf(fptrTask[j], "0\t");

            if(HIST[i*state.num_tasks+j]&OVER) fprintf(fptrTask[j], "%d\t0\t2\t", i);
            else fprintf(fptrTask[j], "%d\t0\t0\t", i);

            if(HIST[i*state.num_tasks+j]&READY) fprintf(fptrTask[j], "%d\t0\t2\n", i);
            else fprintf(fptrTask[j], "%d\t0\t0\n", i);
        }
    }

    for(i=0; i<state.num_tasks; i++)
    {
        fclose(fptrTask[i]);
    }
    
    free(fptrTask);

    fptr=fopen("plot.do", "wt");
    fprintf(fptr, "set term postscript landscape\n");
    fprintf(fptr, "set output \"results.ps\"\n");
    fprintf(fptr, "set xlabel \"Scheduler Tick\" font \"arial, 10\"\n");
    fprintf(fptr, "set xrange[%d:%d]\n", startDisplay, end);
    fprintf(fptr, "set xtics %d, 5, %d font \"arial, 8\"\n", startDisplay, end);
    fprintf(fptr, "set mxtics 5\n");
    fprintf(fptr, "unset ylabel\n");
    fprintf(fptr, "set ytics \"\"\n");
    fprintf(fptr, "unset ytics\n");
    fprintf(fptr, "set yrange [0:3]\n");
    fprintf(fptr, "set key left\n");
    fprintf(fptr, "set style fill solid 1.00 border -1\n");
    fprintf(fptr, "set style arrow 1 backhead filled size screen 0.010,15,25 lt rgb \"#FF0000\"\n");
    fprintf(fptr, "set style arrow 2 head filled size screen 0.010,15,25 lt rgb \"#0000FF\"\n");

    for(j=0; j<state.num_tasks/3; j++) {
      
      fprintf(fptr, "set multiplot layout 3, 1\n");
      for(i=j*3; i<j*3+3; i++)
      {
        fprintf(fptr, "set title \"%s Scheduling (%s)\" font \"arial, 10\"\n", tasks[i].name, state.scheme_name);
        fprintf(fptr, "plot \'TASK-%d.dat' using 1:2 notitle with boxes lt rgb \"%s\" fill, \\\n", tasks[i].ID, GraphColor[i]);
        fprintf(fptr, "    \'\' using 3:4:(0):5 notitle with vectors arrowstyle 1, \\\n");
        fprintf(fptr, "     \'\' using 6:7:(0):8 notitle with vectors arrowstyle 2\n");
      }

      fprintf(fptr, "unset multiplot\n");
      }

    if(state.num_tasks%3>0)
    {
      fprintf(fptr, "set multiplot layout 3, 1\n");
      for(;i<state.num_tasks; i++)
      {
        fprintf(fptr, "set title \"%s Scheduling (%s)\" font \"arial, 10\"\n", tasks[i].name, state.scheme_name);
        fprintf(fptr, "plot \'TASK-%d.dat' using 1:2 notitle with boxes lt rgb \"%s\" fill, \\\n", tasks[i].ID, GraphColor[i]);
        fprintf(fptr, "    \'\' using 3:4:(0):5 notitle with vectors arrowstyle 1, \\\n");
        fprintf(fptr, "     \'\' using 6:7:(0):8 notitle with vectors arrowstyle 2\n");
      }
      fprintf(fptr, "unset multiplot\n");
    }

    fclose(fptr);
    return 0; 
}

